# -*- coding: utf-8 -*-
"""
Board views
"""
from django.views import generic

from djangocodemirror.views import SamplePreviewView, EditorSettingsView

from sveedocuments.models import Page, Insert
from sveedocuments.forms.page import DjangoCodeMirrorSettingsForm

from braces.views import LoginRequiredMixin, PermissionRequiredMixin

class BoardIndexView(LoginRequiredMixin, generic.TemplateView):
    """
    Documents management board
    """
    template_name = "sveedocuments/board/index.html"

    def get(self, request, *args, **kwargs):
        context = {
            'page_count' : Page.objects.count(),
            'page_hidden_count' : Page.objects.filter(visible=False).count(),
            'page_last_revised': Page.objects.all().order_by('-modified')[0:5],
            'insert_count' : Insert.objects.count(),
            'insert_hidden_count' : Insert.objects.filter(visible=False).count(),
        }
        return self.render_to_response(context)

class BoardPagesIndexView(LoginRequiredMixin, generic.TemplateView):
    """
    Board pages index
    """
    template_name = "sveedocuments/board/page_index.html"

    def get(self, request, *args, **kwargs):
        context = {
            'page_list' : Page.objects.filter(),
        }
        return self.render_to_response(context)

class BoardInsertsIndexView(LoginRequiredMixin, generic.TemplateView):
    """
    Board inserts index
    """
    template_name = "sveedocuments/board/insert_index.html"

    def get(self, request, *args, **kwargs):
        context = {
            'insert_list' : Insert.objects.filter().order_by('slug'),
        }
        return self.render_to_response(context)

class PreviewView(LoginRequiredMixin, SamplePreviewView):
    """
    Page preview for a *Page* content
    """
    pass

class BoardEditorSettingsView(LoginRequiredMixin, EditorSettingsView):
    """
    Page form for djangocode-mirror editor settings
    """
    form_class = DjangoCodeMirrorSettingsForm



"""
Asset bundles to use with django-assets
"""
try:
    from django_assets import Bundle, register
except ImportError:
    DJANGO_ASSETS_INSTALLED = False
else:
    DJANGO_ASSETS_INSTALLED = True

    AVALAIBLE_BUNDLES = {
        'sveedocuments_css': Bundle(
            "css/sveedocuments_app.css",
            filters='yui_css',
            output='css/sveedocuments_app.min.css'
        ),
        'sveedocuments_board_js': Bundle(
            "js/jquery/jquery.slugify.js",
            "js/jquery/moment.js",
            "js/jquery/pikaday.js",
            filters='yui_js',
            output='js/sveedocuments_board.min.js'
        ),
    }

    ENABLED_BUNDLES = (
        'sveedocuments_css',
        'sveedocuments_board_js',
    )

    for item in ENABLED_BUNDLES:
        register(item, AVALAIBLE_BUNDLES[item])



# -*- coding: utf-8 -*-
from django.core.cache import cache

from rstview.parser import SourceParser

from sveedocuments.parser import extract_toc

def get_render_with_cache(instance, setting_key="default", force_update_cache=False, initial_header_level=None):
    """
    Get an instance content render by the parser
    
    Use the cache system, content will not be rendered again if it allready exists in the cache
    """
    cache_key = instance.get_render_cache_key(setting=setting_key, header_level=initial_header_level)
    if force_update_cache or not cache.get(cache_key):
        if not instance.content:
            rendered = ''
        rendered = SourceParser(instance.content, setting_key=setting_key, initial_header_level=initial_header_level)
        cache.set(cache_key, rendered)
        return rendered
    return cache.get(cache_key)

def get_toc_with_cache(instance, setting_key="default", force_update_cache=False, initial_header_level=None):
    """
    Extract the TOC from the rendered content of an instance and return it
    
    Use the cache system, content will not be rendered again if it allready exists in the cache
    """
    cache_key = instance.get_toc_cache_key(setting=setting_key, header_level=initial_header_level)
    if force_update_cache or not cache.get(cache_key):
        if not instance.content:
            toc = ''
        toc = extract_toc(instance.content, setting_key=setting_key, initial_header_level=initial_header_level)
        cache.set(cache_key, toc)
        return toc
    return cache.get(cache_key)



# -*- coding: utf-8 -*-
"""
Context processors
"""
from sveedocuments.views.page import PagePDFView

def SveedocumentsContext(request):
    """
    Context processor to add some sveedocuments globals to template context
    """
    return {
        'pdf_export_enabled': not( getattr(PagePDFView, 'is_dummy', False) ),
    }



# -*- coding: utf-8 -*-
"""
App default settings
"""
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

# Add custom cache keys to delete with the command option ``django-admin documents --clearcache``
DOCUMENTS_CACHE_KEYS_TO_CLEAN = getattr(settings, 'DOCUMENTS_CACHE_KEYS_TO_CLEAN', [])

# Forbidden words for slug values in documents
DOCUMENTS_PAGE_RESERVED_SLUGS = getattr(settings, 'DOCUMENTS_PAGE_MENU', ('add','board','help','inserts','preview','sitemap'))

# Default template to generate tree menu
DOCUMENTS_PAGE_TREEMENU = getattr(settings, 'DOCUMENTS_PAGE_TREEMENU', "sveedocuments/menu_tree.html")
# Default template to generate flat menu
DOCUMENTS_PAGE_FLATMENU = getattr(settings, 'DOCUMENTS_PAGE_FLATMENU', "sveedocuments/menu_flat.html")

# Enable or disable Pages archiving
DOCUMENTS_PAGE_ARCHIVED = True

# Available templates for Pages
DOCUMENTS_PAGE_TEMPLATES = {
    'default': ('sveedocuments/page_details/page_default.html', _('Default template with document content only')),
    # Disabled, because TOC extract is buggy
    #'with_toc': ('sveedocuments/page_details/page_with-toc.html', _('Template with a Table Of Content and the document content')),
}
DOCUMENTS_PAGE_TEMPLATES.update(getattr(settings, 'DOCUMENTS_PAGE_TEMPLATES', {}))

DOCUMENTS_PAGE_TEMPLATE_DEFAULT = getattr(settings, 'DOCUMENTS_PAGE_TEMPLATE_DEFAULT', 'elastic')

# Active a silent reporter to avoid warnings about missing page on role ``:page:xxx``
# If ``True`` the role will be transformed to link despite the page does not exist, 
# if ``False`` the warning will be inserted in the render
DOCUMENTS_PARSER_WIKIROLE_SILENT_WARNING = getattr(settings, 'DOCUMENTS_PARSER_WIKIROLE_SILENT_WARNING', False)

# This is the preface that will be added at start of the Page content when they are 
# exported to PDF
# The default preface put some document informations, add summary of content on the 
# first page and active the section numbering.
_PDF_PREFACE = u"""==================
{page.title}
==================

:author: {page.author.first_name} {page.author.last_name} <{page.author.email}>
:date: {page.modified}
:revision: $LastChangedRevision: {page.current_revision} $

.. class:: alert alert-info pull-right

.. contents::

.. section-numbering::

.. raw:: pdf

   PageBreak oneColumn

"""
DOCUMENTS_EXPORT_PDF_PREFACE = getattr(settings, 'DOCUMENTS_EXPORT_PDF_PREFACE', _PDF_PREFACE)
DOCUMENTS_EXPORT_PDF_HEADER = getattr(settings, 'DOCUMENTS_EXPORT_PDF_PREFACE', "###Section###")
DOCUMENTS_EXPORT_PDF_FOOTER = getattr(settings, 'DOCUMENTS_EXPORT_PDF_PREFACE', "Page ###Page### / ###Total###")

"""
Internal settings only, can't be overriden from your webapp settings
"""

# Cache keys for document elements
PAGE_RENDER_CACHE_KEY_NAME = 'documents-render-page_{id}-setting_{setting}'
PAGEREV_RENDER_CACHE_KEY_NAME = 'documents-render-page-revision_{id}-setting_{setting}'
INSERT_RENDER_CACHE_KEY_NAME = 'documents-render-insert_{id}-setting_{setting}-hlv_{header_level}'
PAGE_TOC_CACHE_KEY_NAME = 'documents-toc-page_{id}-setting_{setting}'
PAGEREV_TOC_CACHE_KEY_NAME = 'documents-toc-page-revision_{id}-setting_{setting}'
INSERT_TOC_CACHE_KEY_NAME = 'documents-toc-insert_{id}-setting_{setting}-hlv_{header_level}'
PAGE_SLUGS_CACHE_KEY_NAME = 'documents-page_slugs'
PAGE_ATTACHMENTS_SLUGS_CACHE_KEY_NAME = 'documents-page-attachments-slugs'

"""
WARNING: Sample additional Django-CodeMirror settings, you have to put them yourself in your project settings
"""
DJANGOCODEMIRROR_USER_SETTINGS_COOKIE_NAME = "djangocodemirror_user_settings"
CODEMIRROR_SETTINGS = {
    'sveetchies-documents-edit-page': {
        'mode': 'rst',
        'csrf': 'CSRFpass',
        'preview_url': ('sveedocuments:preview',),
        'quicksave_url': ('sveedocuments:page-quicksave',),
        'quicksave_datas': 'DJANGOCODEMIRROR_OBJECT',
        'lineWrapping': True,
        'lineNumbers': True,
        'search_enabled': True,
        'settings_cookie': DJANGOCODEMIRROR_USER_SETTINGS_COOKIE_NAME,
        'help_link': ('sveedocuments:help',),
        'settings_url': ('sveedocuments:editor-settings', [], {}),
    },
    'sveetchies-documents-edit-insert': {
        'mode': 'rst',
        'csrf': 'CSRFpass',
        'preview_url': ('sveedocuments:preview',),
        'quicksave_url': ('sveedocuments:insert-quicksave',),
        'quicksave_datas': 'DJANGOCODEMIRROR_OBJECT',
        'lineWrapping': True,
        'lineNumbers': True,
        'search_enabled': True,
        'settings_cookie': DJANGOCODEMIRROR_USER_SETTINGS_COOKIE_NAME,
        'help_link': ('sveedocuments:help',),
        'settings_url': ('sveedocuments:editor-settings', [], {}),
    },
}
CODEMIRROR_SETTINGS['sveetchies-documents-add-page'] = CODEMIRROR_SETTINGS['sveetchies-documents-edit-page'].copy()
CODEMIRROR_SETTINGS['sveetchies-documents-add-page']['quicksave_url'] = None
CODEMIRROR_SETTINGS['sveetchies-documents-add-insert'] = CODEMIRROR_SETTINGS['sveetchies-documents-edit-insert'].copy()
CODEMIRROR_SETTINGS['sveetchies-documents-add-insert']['quicksave_url'] = None



# -*- coding: utf-8 -*-
"""
Url's map for documents board
"""
from django.conf.urls import *

from sveedocuments.views.board import (BoardIndexView, PreviewView, BoardEditorSettingsView,
                                        BoardPagesIndexView, BoardInsertsIndexView)
from sveedocuments.views.page import (PageCreateView, PageQuicksaveView,
                                      PageEditView, PageDeleteView, PageDeleteView, PageHistoryView, 
                                      PageAttachmentsView, PageAttachmentDeleteView)
from sveedocuments.views.insert import (InsertCreateView, InsertEditView, 
                                        InsertDeleteView, InsertQuicksaveView)

urlpatterns = patterns('',
    url(r'^$', BoardIndexView.as_view(), name='board'),
    
    url(r'^preview/$', PreviewView.as_view(), name='preview'),
    url(r'^editor_settings/$', BoardEditorSettingsView.as_view(), name='editor-settings'),
    
    url(r'^quicksave/page/$', PageQuicksaveView.as_view(), name='page-quicksave'),
    url(r'^quicksave/insert/$', InsertQuicksaveView.as_view(), name='insert-quicksave'),
    
    url(r'^pages/$', BoardPagesIndexView.as_view(), name='page-index'),
    url(r'^pages/add/$', PageCreateView.as_view(), name='page-add'),
    url(r'^pages/(?P<slug>[-\w]+)/add/$', PageCreateView.as_view(), name='page-add-child'),
    url(r'^pages/(?P<slug>[-\w]+)/delete/$', PageDeleteView.as_view(), name='page-delete'),
    url(r'^pages/(?P<slug>[-\w]+)/edit/$', PageEditView.as_view(), name='page-edit'),
    url(r'^pages/(?P<slug>[-\w]+)/edit/history/$', PageHistoryView.as_view(), name='page-history'),
    url(r'^pages/(?P<slug>[-\w]+)/edit/attachments/$', PageAttachmentsView.as_view(), name='page-attachments'),
    url(r'^pages/(?P<slug>[-\w]+)/edit/attachments/(?P<pk>\d+)/delete/$', PageAttachmentDeleteView.as_view(), name='page-attachments-delete'),
    
    url(r'^inserts/$', BoardInsertsIndexView.as_view(), name='insert-index'),
    url(r'^inserts/add/$', InsertCreateView.as_view(), name='insert-add'),
    url(r'^inserts/(?P<slug>[-\w]+)/delete/$', InsertDeleteView.as_view(), name='insert-delete'),
    url(r'^inserts/(?P<slug>[-\w]+)/edit/$', InsertEditView.as_view(), name='insert-edit'),
)



# -*- coding: utf-8 -*-
"""
Templates tags divers pour les documents
"""
from django import template
from django.utils.safestring import mark_safe

from sveedocuments.models import Insert, Page
from sveedocuments.templatetags import get_render_with_cache, get_toc_with_cache
from sveedocuments.utils.templatetags import resolve_string_or_variable

register = template.Library()

class DocumentTagContext(template.Node):
    """
    Generate the document elements (content, TOC, navigation)
    
    This doesn't add any HTML, all elements are injected in the current template context 
    where you can use them.
    """
    def __init__(self, insert_instance_varname, setting_key_varname=None, title_level_varname=None, with_toc=True, with_navigation=True):
        """
        :type insert_instance_varname: string or object ``django.db.models.Model``
        :param insert_instance_varname: Nom de variable de l'instance ou un string 
                                        pour le slug.
                                        
        :type setting_key_varname: string
        :param setting_key_varname: (optionnel) Nom clé du schéma d'options du parser à 
                                    utiliser, par défaut utilise le schéma "short" dédié 
                                    aux documents à insérer.
        
        :type title_level_varname: int
        :param title_level_varname: Niveau minimale des titres de premier niveau, par 
                                    défaut utilise celui du schéma. Si une valeur est 
                                    spécifiée, elle sera utilisée comme limite, par 
                                    exemple avec une valeur de 4, aucun titres ne sera 
                                    plus haut qu'un ``<h4/>`` (donc seulement h4, h5, h6).
        """
        self.insert_instance_varname = insert_instance_varname
        self.setting_key_varname = setting_key_varname
        self.title_level_varname = title_level_varname
        self.with_toc = with_toc
        self.with_navigation = with_navigation
    
    def render(self, context):
        """
        TODO: Lever une exception de template évident quand le setting demandé n'existe 
              pas.
        
        :type context: object ``django.template.Context``
        :param context: Objet du contexte du tag.
        
        :rtype: string
        :return: Le rendu généré pour le tag capturé.
        """
        html = ''
        parser_kwargs = {}
        content_render = toc_render = navigation_render = None
        
        # Résolution des arguments
        instance = resolve_string_or_variable(self.insert_instance_varname, context)
        
        setting_key = resolve_string_or_variable(self.setting_key_varname, context)
        if setting_key:
            parser_kwargs['setting_key'] = setting_key
        
        title_level = resolve_string_or_variable(self.title_level_varname, context)
        if title_level:
            parser_kwargs['initial_header_level'] = title_level
        
        content_render = mark_safe( get_render_with_cache(instance, **parser_kwargs) )
        #if self.with_toc:
            #toc_render = mark_safe( get_toc_with_cache(instance, **parser_kwargs) )
        if self.with_navigation and isinstance(instance, Page):
            navigation = instance.get_descendants(include_self=False).filter(visible=True)
        
        context.update({
            #'document_toc': toc_render,
            'document_navigation': navigation,
            'document_render': content_render,
        })
        
        return ''

class InsertTagRender(template.Node):
    """
    Generate HTML of the node *InsertTagRender*
    """
    def __init__(self, insert_instance_varname, setting_key_varname=None, title_level_varname=None):
        """
        :type insert_instance_varname: string or object ``django.db.models.Model``
        :param insert_instance_varname: Nom de variable de l'instance ou un string 
                                        pour le slug.
                                        
        :type setting_key_varname: string
        :param setting_key_varname: (optionnel) Nom clé du schéma d'options du parser à 
                                    utiliser, par défaut utilise le schéma "short" dédié 
                                    aux documents à insérer.
        
        :type title_level_varname: int
        :param title_level_varname: Niveau minimale des titres de premier niveau, par 
                                    défaut utilise celui du schéma. Si une valeur est 
                                    spécifiée, elle sera utilisée comme limite, par 
                                    exemple avec une valeur de 4, aucun titres ne sera 
                                    plus haut qu'un ``<h4/>`` (donc seulement h4, h5, h6).
        """
        self.insert_instance_varname = insert_instance_varname
        self.setting_key_varname = setting_key_varname
        self.title_level_varname = title_level_varname
    
    def render(self, context):
        """
        TODO: Lever une exception de template évident quand le setting demandé n'existe 
              pas.
        
        :type context: object ``django.template.Context``
        :param context: Objet du contexte du tag.
        
        :rtype: string
        :return: Le rendu généré pour le tag capturé.
        """
        html = ''
        parser_kwargs = {}
        
        # Résolution de l'instance du formulaire
        self.insert_instance = resolve_string_or_variable(self.insert_instance_varname, context)
        
        self.setting_key = resolve_string_or_variable(self.setting_key_varname, context)
        if self.setting_key:
            parser_kwargs['setting_key'] = self.setting_key
        
        self.title_level = resolve_string_or_variable(self.title_level_varname, context)
        if self.title_level:
            parser_kwargs['initial_header_level'] = self.title_level
        
        # Réception d'un slug pour récupérer l'instance
        if isinstance(self.insert_instance, basestring):
            try:
                instance = Insert.objects.get(slug=self.insert_instance, visible=True)
            except Insert.DoesNotExist:
                return ''
            else:
                self.insert_instance = instance
        
        html = get_render_with_cache(self.insert_instance, **parser_kwargs)
        
        return mark_safe(html)

@register.tag(name="document_context")
def do_document_context(parser, token):
    """
    Push the document elements in the context
    
    Arguments :
    
    document_object
        Instance (``Page`` ou ``Insert``) ou un string contenant un slug de l'instance 
        à récupérer
        
    Exemple d'utilisation par une instance Insert : ::
    
        {% document_insert instance %}
        
    Exemple d'utilisation par le slug de l'Insert à utiliser : ::
    
        {% document_insert "my-insert-slug" %}
    
    :type parser: object ``django.template.Parser``
    :param parser: Objet du parser de template.
    
    :type token: object ``django.template.Token``
    :param token: Objet de la chaîne découpée du tag capturé dans le template.
    
    :rtype: object ``InsertTagRender``
    :return: L'objet du générateur de rendu du tag.
    """
    args = token.split_contents()
    if len(args) < 2:
        raise template.TemplateSyntaxError, "You need to specify at less an instance (Insert or Page) or a slug"
    else:
        return DocumentTagContext(*args[1:])

@register.tag(name="document_insert")
def do_document_insert(parser, token):
    """
    Display the parsed content of an **Insert** document
    
    Arguments :
    
    document_object
        Instance "Insert" ou un string contenant un slug de l'instance à récupérer
        
    Exemple d'utilisation par une instance Insert : ::
    
        {% document_insert instance %}
        
    Exemple d'utilisation par le slug de l'Insert à utiliser : ::
    
        {% document_insert "my-insert-slug" %}
    
    :type parser: object ``django.template.Parser``
    :param parser: Objet du parser de template.
    
    :type token: object ``django.template.Token``
    :param token: Objet de la chaîne découpée du tag capturé dans le template.
    
    :rtype: object ``InsertTagRender``
    :return: L'objet du générateur de rendu du tag.
    """
    args = token.split_contents()
    if len(args) < 2:
        raise template.TemplateSyntaxError, "You need to specify at less an \"Insert\" instance or a slug"
    else:
        return InsertTagRender(*args[1:])



# -*- coding: utf-8 -*-
"""
Page forms
"""
from django import forms
from django.utils.translation import ugettext as _

from mptt.forms import TreeNodeChoiceField

from djangocodemirror.fields import CodeMirrorWidget
from djangocodemirror.forms import DjangoCodeMirrorSettingsForm as DjangoCodeMirrorSettingsFormBase

from rstview.parser import SourceReporter, map_parsing_errors

from sveedocuments.local_settings import DOCUMENTS_PAGE_RESERVED_SLUGS
from sveedocuments.models import Page, documents_page_update_signal
from sveedocuments.forms import CrispyFormMixin

class PageForm(CrispyFormMixin, forms.ModelForm):
    """
    Page form
    """
    crispy_form_helper_path = 'sveedocuments.forms.crispies.page_helper'
    codemirror_config_name = 'sveetchies-documents-add-page'
    
    def __init__(self, author=None, parent=None, *args, **kwargs):
        self.author = author
        self.parent = parent

        self.crispy_form_helper_kwargs = {'parent': self.parent}
        
        super(PageForm, self).__init__(*args, **kwargs)
        super(forms.ModelForm, self).__init__(*args, **kwargs)
        
        # En cas d'édition, limite les choix de parenté à tout ceux qui ne sont pas 
        # descendant de l'instance pour empêcher une exception 
        # sur une erreur de déplacement
        extra_settings = {}
        if self.instance.id:
            parent_queryset = Page.objects.all()
            children = self.instance.get_descendants(include_self=True).values_list('id', flat=True)
            parent_queryset = parent_queryset.exclude(id__in=children)
            self.fields['parent'] = TreeNodeChoiceField(queryset=parent_queryset, empty_label=_(u"-- Root --"), required=False)
        # En cas de création, mode pour ajouter une page directement sous un "parent"
        else:
            if self.parent:
                del self.fields['parent']
            # Désactive l'option de sauvegarde rapide
            extra_settings = {'quicksave_url': None}
            
        # Widget d'édition du contenu
        self.fields['content'].widget = CodeMirrorWidget(attrs={'rows': 30}, config_name=self.codemirror_config_name)
    
    def clean_slug(self):
        slug = self.cleaned_data.get("slug")
        if slug:
            if slug in DOCUMENTS_PAGE_RESERVED_SLUGS:
                raise forms.ValidationError(_('This is a reserved keyword'))
        return slug
    
    def clean_content(self):
        """
        Parse le contenu pour vérifier qu'il ne contient par d'erreurs de syntaxe
        """
        content = self.cleaned_data.get("content")
        if content:
            errors = SourceReporter(content)
            if errors:
                raise forms.ValidationError(map(map_parsing_errors, errors))
        return content
    
    def save(self, *args, **kwargs):
        edited = (self.instance and self.instance.id)

        instance = super(PageForm, self).save(commit=False, *args, **kwargs)
        instance.author = self.author
        if self.parent:
            instance.parent = self.parent
        instance.save()
        
        # Send a Django signal for page update
        documents_page_update_signal.send(sender=self, page_instance=instance, edited=edited)
        
        return instance
    
    class Meta:
        model = Page
        exclude = ('created', 'author')
        widgets = {
            'published': forms.SplitDateTimeWidget,
        }


class PageEditForm(PageForm):
    crispy_form_helper_path = 'sveedocuments.forms.crispies.page_edit_helper'
    codemirror_config_name = 'sveetchies-documents-edit-page'


class PageQuickForm(forms.ModelForm):
    """
    *Page* quicksave form
    
    This is only about the instance *content* field
    
    TODO: This should also accepts the associated 'comment' field, but in first the 
          editor must implement to send the "comment" content with the "content" 
          content.
    """
    def __init__(self, author=None, *args, **kwargs):
        self.author = author
        super(PageQuickForm, self).__init__(*args, **kwargs)
    
    def clean_content(self):
        """
        Parse content to check syntax alerts
        """
        content = self.cleaned_data.get("content")
        if content:
            errors = SourceReporter(content)
            if errors:
                raise forms.ValidationError(map(map_parsing_errors, errors))
        return content
    
    def save(self, *args, **kwargs):
        instance = super(PageQuickForm, self).save(commit=True, *args, **kwargs)
        return instance
    
    class Meta:
        model = Page
        fields = ('content',)

class DjangoCodeMirrorSettingsForm(DjangoCodeMirrorSettingsFormBase):
    editor_settings_url = 'sveedocuments:editor-settings'



# -*- coding: utf-8 -*-
"""
Utilities
"""
import copy, itertools

def _get_cache_keyset(keytpl, **kwargs):
    """
    Return a list of possible cache keys of a document (*Pages* or *Insert*)
    
    Because a document can have various active cache key for his various possible render 
    (from the parser settings) and elements, this is useful to invalidate caches after 
    changes.
    
    >>> _get_cache_keyset("id_{id}", **{'id':42})
    >>> ['id_42']
    >>> _get_cache_keyset("id_{id}-set_{set}-hlv_{hlv}", **{'id':42, 'setting': ['short','long'], 'hlv': range(1, 4)})
    >>> ['id_42-setting_short-hlv_1', 'id_42-setting_long-hlv_1', 'id_42-setting_short-hlv_2', 'id_42-setting_long-hlv_2', 'id_42-setting_short-hlv_3', 'id_42-setting_long-hlv_3']
    """
    keys = []
    # Tout les arguments qui sont des list/tuple
    list_args = filter((lambda keyname: isinstance(kwargs[keyname], list) or isinstance(kwargs[keyname], tuple)), kwargs.keys())
    # Tout les arguments qui ne sont pas list/tuple donc des string/int, donc des valeurs uniques
    nonlist_args = [k for k in kwargs if k not in list_args]
    # Context de base avec les arguments 'simples'
    base_context = dict(zip(nonlist_args, [kwargs[k] for k in nonlist_args]))
    
    # Produit une liste de toute les possibilités des arguments
    for item in itertools.product(*[kwargs[item] for item in list_args]):
        context = base_context.copy()
        context.update(dict(zip(list_args, item)))
        keys.append(keytpl.format(**context))

    return keys




"""
Common forum mixins
"""
from django.conf import settings
from django.shortcuts import render_to_response
from django.template import RequestContext, TemplateDoesNotExist

from guardian.conf import settings as guardian_settings
from guardian.mixins import PermissionRequiredMixin

class PageChangePermissionRequiredMixin(PermissionRequiredMixin):
    """
    Mixin to check for "page_change" permission with a Page object
    
    This used for views where self.object is not a Page instance.
    
    The page instance must be set in a "page_object" View attribute.
    """
    permission_required = 'sveedocuments.change_page'
    
    def check_permissions(self, request):
        perms = self.get_required_permissions(request)
        page_instance = (hasattr(self, 'get_page_object') and self.get_page_object()
            or getattr(self, 'page_object', None))
        has_permissions = False
        # global perms check first (if accept_global_perms)
        has_permissions = any(request.user.has_perm(perm) for perm in perms)
        # if still no permission granted, try obj perms
        if not has_permissions:
            has_permissions = any(request.user.has_perm(perm, page_instance) for perm in perms)
        
        # Raise a forbidden response if still no permission at all
        if not has_permissions:
            try:
                response = render_to_response(getattr(guardian_settings, 'TEMPLATE_403', '403.html'), {}, RequestContext(request))
                response.status_code = 403
                return response
            except TemplateDoesNotExist as e:
                if settings.DEBUG:
                    raise e
        return False



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Insert'
        db.create_table(u'sveedocuments_insert', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
            ('author', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'])),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('slug', self.gf('django.db.models.fields.SlugField')(unique=True, max_length=75)),
            ('visible', self.gf('django.db.models.fields.BooleanField')(default=True)),
            ('content', self.gf('django.db.models.fields.TextField')()),
        ))
        db.send_create_signal(u'sveedocuments', ['Insert'])

        # Adding model 'Page'
        db.create_table(u'sveedocuments_page', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(blank=True)),
            ('author', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'])),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('published', self.gf('django.db.models.fields.DateTimeField')(blank=True)),
            ('template', self.gf('django.db.models.fields.CharField')(default='elastic', max_length=50)),
            ('order', self.gf('django.db.models.fields.SmallIntegerField')(default=1)),
            ('visible', self.gf('django.db.models.fields.BooleanField')(default=True)),
            ('content', self.gf('django.db.models.fields.TextField')()),
            ('comment', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
            ('slug', self.gf('django.db.models.fields.SlugField')(unique=True, max_length=75)),
            ('parent', self.gf('mptt.fields.TreeForeignKey')(blank=True, related_name=u'sveedocuments_page_children', null=True, to=orm['sveedocuments.Page'])),
            (u'lft', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
            (u'rght', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
            (u'tree_id', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
            (u'level', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
        ))
        db.send_create_signal(u'sveedocuments', ['Page'])

        # Adding model 'PageRevision'
        db.create_table(u'sveedocuments_pagerevision', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(blank=True)),
            ('author', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'])),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('published', self.gf('django.db.models.fields.DateTimeField')(blank=True)),
            ('template', self.gf('django.db.models.fields.CharField')(default='elastic', max_length=50)),
            ('order', self.gf('django.db.models.fields.SmallIntegerField')(default=1)),
            ('visible', self.gf('django.db.models.fields.BooleanField')(default=True)),
            ('content', self.gf('django.db.models.fields.TextField')()),
            ('comment', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('slug', self.gf('django.db.models.fields.SlugField')(max_length=75)),
            ('page', self.gf('django.db.models.fields.related.ForeignKey')(related_name='revision', to=orm['sveedocuments.Page'])),
            ('parent', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='revision_parent', null=True, to=orm['sveedocuments.Page'])),
            ('revision', self.gf('django.db.models.fields.IntegerField')(default=0)),
        ))
        db.send_create_signal(u'sveedocuments', ['PageRevision'])

        # Adding model 'Attachment'
        db.create_table(u'sveedocuments_attachment', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('page', self.gf('django.db.models.fields.related.ForeignKey')(related_name='attachment', to=orm['sveedocuments.Page'])),
            ('author', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'])),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=75, blank=True)),
            ('slug', self.gf('django.db.models.fields.CharField')(max_length=75, blank=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('file', self.gf('django.db.models.fields.files.FileField')(max_length=255)),
            ('size', self.gf('django.db.models.fields.IntegerField')(default=0)),
            ('content_type', self.gf('django.db.models.fields.CharField')(max_length=120, null=True)),
        ))
        db.send_create_signal(u'sveedocuments', ['Attachment'])


    def backwards(self, orm):
        # Deleting model 'Insert'
        db.delete_table(u'sveedocuments_insert')

        # Deleting model 'Page'
        db.delete_table(u'sveedocuments_page')

        # Deleting model 'PageRevision'
        db.delete_table(u'sveedocuments_pagerevision')

        # Deleting model 'Attachment'
        db.delete_table(u'sveedocuments_attachment')


    models = {
        u'auth.group': {
            'Meta': {'object_name': 'Group'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        u'auth.permission': {
            'Meta': {'ordering': "(u'content_type__app_label', u'content_type__model', u'codename')", 'unique_together': "((u'content_type', u'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        u'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'sveedocuments.attachment': {
            'Meta': {'object_name': 'Attachment'},
            'author': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['auth.User']"}),
            'content_type': ('django.db.models.fields.CharField', [], {'max_length': '120', 'null': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'file': ('django.db.models.fields.files.FileField', [], {'max_length': '255'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'page': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'attachment'", 'to': u"orm['sveedocuments.Page']"}),
            'size': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'slug': ('django.db.models.fields.CharField', [], {'max_length': '75', 'blank': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '75', 'blank': 'True'})
        },
        u'sveedocuments.insert': {
            'Meta': {'object_name': 'Insert'},
            'author': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['auth.User']"}),
            'content': ('django.db.models.fields.TextField', [], {}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'slug': ('django.db.models.fields.SlugField', [], {'unique': 'True', 'max_length': '75'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'visible': ('django.db.models.fields.BooleanField', [], {'default': 'True'})
        },
        u'sveedocuments.page': {
            'Meta': {'object_name': 'Page'},
            'author': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['auth.User']"}),
            'comment': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'content': ('django.db.models.fields.TextField', [], {}),
            'created': ('django.db.models.fields.DateTimeField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            u'level': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            u'lft': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.SmallIntegerField', [], {'default': '1'}),
            'parent': ('mptt.fields.TreeForeignKey', [], {'blank': 'True', 'related_name': "u'sveedocuments_page_children'", 'null': 'True', 'to': u"orm['sveedocuments.Page']"}),
            'published': ('django.db.models.fields.DateTimeField', [], {'blank': 'True'}),
            u'rght': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'slug': ('django.db.models.fields.SlugField', [], {'unique': 'True', 'max_length': '75'}),
            'template': ('django.db.models.fields.CharField', [], {'default': "'elastic'", 'max_length': '50'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            u'tree_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'visible': ('django.db.models.fields.BooleanField', [], {'default': 'True'})
        },
        u'sveedocuments.pagerevision': {
            'Meta': {'object_name': 'PageRevision'},
            'author': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['auth.User']"}),
            'comment': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'content': ('django.db.models.fields.TextField', [], {}),
            'created': ('django.db.models.fields.DateTimeField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'order': ('django.db.models.fields.SmallIntegerField', [], {'default': '1'}),
            'page': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'revision'", 'to': u"orm['sveedocuments.Page']"}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'revision_parent'", 'null': 'True', 'to': u"orm['sveedocuments.Page']"}),
            'published': ('django.db.models.fields.DateTimeField', [], {'blank': 'True'}),
            'revision': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '75'}),
            'template': ('django.db.models.fields.CharField', [], {'default': "'elastic'", 'max_length': '50'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'visible': ('django.db.models.fields.BooleanField', [], {'default': 'True'})
        }
    }

    complete_apps = ['sveedocuments']


# -*- coding: utf-8 -*-
import warnings

from django.utils.importlib import import_module

def safe_import_module(path, default=None):
    """
    Try to import the specified module from the given Python path
    
    @path is a string containing a Python path to the wanted module, @default is 
    an object to return if import fails, it can be None, a callable or whatever you need.
    
    Return a object or None
    """
    if path is None:
        return default
    
    dot = path.rindex('.')
    module_name = path[:dot]
    class_name = path[dot + 1:]
    try:
        _class = getattr(import_module(module_name), class_name)
        return _class
    except (ImportError, AttributeError):
        warnings.warn('%s cannot be imported' % path, RuntimeWarning)
    return default



from setuptools import setup, find_packages

setup(
    name='sveedocuments',
    version=__import__('sveedocuments').__version__,
    description=__import__('sveedocuments').__doc__,
    long_description=open('README.rst').read(),
    author='David Thenon',
    author_email='sveetch@gmail.com',
    url='http://pypi.python.org/pypi/sveedocuments',
    license='MIT',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        "Development Status :: 5 - Production/Stable",
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Text Processing :: Markup',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'docutils>=0.7',
        'rstview>=0.2',
        'autobreadcrumbs>=1.0',
        'djangocodemirror>=0.9.6',
        'django-mptt>=0.5.2',
        'crispy-forms-foundation>=0.3.4',
        'django-guardian>=1.2.0',
        'django-braces>=1.3.0',
    ],
    include_package_data=True,
    zip_safe=False
)


"""
Crispy forms layouts
"""
from django import forms
from django.utils.translation import ugettext as _

from crispy_forms.helper import FormHelper
from crispy_forms_foundation.layout import Layout, Fieldset, SplitDateTimeField, RowFluid, Column, ButtonHolder, Submit

def page_helper(form_tag=True, parent=None, edit_mode=False):
    """
    Page's form layout helper
    """
    helper = FormHelper()
    helper.form_action = '.'
    helper.attrs = {'data_abide': ''}
    helper.form_tag = form_tag
    
    # Put "parent" field in the layout if it is not forced in kwargs
    layout_fields_publish = [
        RowFluid(
            Column(SplitDateTimeField('published'), css_class='small-12 medium-6'),
            Column('slug', css_class='small-12 medium-6'),
        )
    ]
    if not parent:
        layout_fields_publish = ['parent']+layout_fields_publish
    
    helper.layout = Layout(
        Fieldset(
            _('Content'),
            'title',
            'content',
        ),
        Fieldset(
            _('Display settings'),
            RowFluid(
                Column('template', css_class='small-12 medium-6'),
                Column('order', css_class='small-12 medium-3'),
                Column('visible', css_class='small-12 medium-3'),
            ),
        ),
        Fieldset(
            _('Publish settings'),
            *layout_fields_publish
        ),
        Fieldset(
            _('History'),
            'comment',
        ),
        ButtonHolder(
            Submit('submit_and_continue', _('Save and continue')),
            Submit('submit', _('Save')),
        ),
    )
    
    return helper

def page_edit_helper(form_tag=True, parent=None):
    return page_helper(form_tag=form_tag, parent=parent, edit_mode=True)


def insert_helper(form_tag=True, edit_mode=False):
    """
    Insert's form layout helper
    """
    helper = FormHelper()
    helper.form_action = '.'
    helper.attrs = {'data_abide': ''}
    helper.form_tag = form_tag
    
    helper.layout = Layout(
        Fieldset(
            _('Content'),
            'title',
            'content',
        ),
        Fieldset(
            _('Display settings'),
            RowFluid(
                Column('slug', css_class='small-12 medium-9'),
                Column('visible', css_class='small-12 medium-3'),
            ),
        ),
        ButtonHolder(
            Submit('submit_and_continue', _('Save and continue')),
            Submit('submit', _('Save')),
        ),
    )
    
    return helper

def insert_edit_helper(form_tag=True):
    return insert_helper(form_tag=form_tag, edit_mode=True)


def attachment_helper(form_tag=True, edit_mode=False):
    """
    Attachment's form layout helper
    """
    helper = FormHelper()
    helper.form_action = '.'
    helper.attrs = {'data_abide': ''}
    helper.form_tag = form_tag
    
    helper.layout = Layout(
        RowFluid(
            Column('title', css_class='small-12 medium-4'),
            Column('slug', css_class='small-12 medium-4'),
            Column('file', css_class='small-12 medium-4'),
        ),
        ButtonHolder(
            Submit('submit', _('Save')),
        ),
    )
    
    return helper

def attachment_edit_helper(form_tag=True):
    return attachment_helper(form_tag=form_tag, edit_mode=True)



# -*- coding: utf-8 -*-
"""
Forms
"""
from django import forms
from django.utils.translation import ugettext as _

from sveedocuments.models import Attachment
from sveedocuments.forms import CrispyFormMixin

class AttachmentForm(CrispyFormMixin, forms.ModelForm):
    """
    Attachment form
    """
    crispy_form_helper_path = 'sveedocuments.forms.crispies.attachment_helper'
    
    def __init__(self, *args, **kwargs):
        self.page = kwargs.pop('page_instance')
        self.author = kwargs.pop('author')
        
        super(AttachmentForm, self).__init__(*args, **kwargs)
        super(forms.ModelForm, self).__init__(*args, **kwargs)
    
    def clean_file(self):
        file = self.cleaned_data['file']
        if file:
            self.size = file._size
            self.content_type = file.content_type

        return file
    
    def save(self, *args, **kwargs):
        instance = super(AttachmentForm, self).save(commit=False, *args, **kwargs)
        instance.page = self.page
        instance.author = self.author
        instance.size = self.size
        instance.content_type = self.content_type
        instance.save()
        
        return instance
    
    class Meta:
        model = Attachment
        fields = ('title', 'slug', 'file')
